use std::any::TypeId;
use std::io;
use std::io::BufRead;

use serde::de::DeserializeOwned;
use thiserror::Error;

pub use deckster_shared as shared;
use deckster_shared::handler_communication::{
    HandlerCommand, HandlerEvent, HandlerInitializationError, HandlerInitializationResultMessage, InitialHandlerMessage,
};

#[derive(Debug, Error)]
pub enum RunError {
    #[error("A stdin line could not be read.")]
    LineIo(#[from] io::Error),

    #[error("A stdin line could not be deserialized: {description}")]
    LineDeserialization { line: String, description: String },
}

pub trait DecksterHandler {
    fn handle(&mut self, event: HandlerEvent);
}

enum Stage<I, H: DecksterHandler> {
    Initialization(I),
    Active(H),
}

pub fn run<
    GlobalConfig: Clone + DeserializeOwned + 'static,
    KeyConfig: Clone + DeserializeOwned + 'static,
    KnobConfig: Clone + DeserializeOwned + 'static,
    H: DecksterHandler,
    I: FnOnce(InitialHandlerMessage<GlobalConfig, KeyConfig, KnobConfig>) -> Result<H, HandlerInitializationError>,
>(
    init_handler: I,
) -> Result<(), RunError> {
    let mut stage: Stage<I, H> = Stage::Initialization(init_handler);

    let requires_global_config = TypeId::of::<GlobalConfig>() != TypeId::of::<()>();
    let supports_keys = TypeId::of::<KeyConfig>() != TypeId::of::<()>();
    let supports_knobs = TypeId::of::<KnobConfig>() != TypeId::of::<()>();

    let handle = io::stdin().lock();
    for line in handle.lines() {
        let line = line.map_err(RunError::LineIo)?;

        match stage {
            Stage::Initialization(init_handler) => {
                // TODO: Serialize the global config and each key/knob config separately for more specific error messages.
                let initial_message = serde_json::from_str::<InitialHandlerMessage<GlobalConfig, KeyConfig, KnobConfig>>(&line);

                match initial_message {
                    Ok(initial_message) => match init_handler(initial_message) {
                        Ok(h) => {
                            println!(
                                "{}",
                                serde_json::to_string(&HandlerInitializationResultMessage::Ready).expect("serialization of a known value always works")
                            );
                            stage = Stage::Active(h)
                        }
                        Err(error) => {
                            println!(
                                "{}",
                                serde_json::to_string(&HandlerInitializationResultMessage::Error { error }).expect("no reason to fail")
                            );
                            return Ok(());
                        }
                    },
                    Err(err) => {
                        println!(
                            "{}",
                            serde_json::to_string(&HandlerInitializationResultMessage::Error {
                                error: HandlerInitializationError::InvalidConfig {
                                    requires_global_config,
                                    supports_keys,
                                    supports_knobs,
                                    message: err.to_string().into_boxed_str(),
                                }
                            })
                            .expect("no reason to fail")
                        );
                        return Ok(());
                    }
                }
            }
            Stage::Active(mut h) => {
                let event: HandlerEvent = serde_json::from_str(&line).map_err(|e| RunError::LineDeserialization {
                    line,
                    description: e.to_string(),
                })?;

                let should_stop = matches!(event, HandlerEvent::Stop);

                h.handle(event);
                stage = Stage::Active(h);

                if should_stop {
                    break;
                }
            }
        }
    }

    Ok(())
}

pub fn send_command(command: HandlerCommand) {
    println!("{}", serde_json::to_string(&command).expect("no reason to fail"));
}
